'use strict';

exports.__esModule = true;
exports.fontFace = exports.inserted = exports.sheet = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

exports.flush = flush;
exports.css = css;
exports.injectGlobal = injectGlobal;
exports.keyframes = keyframes;
exports.hydrate = hydrate;
exports.objStyle = objStyle;

var _sheet = require('./sheet');

var _hash = require('./hash');

var sheet = exports.sheet = new _sheet.StyleSheet();
sheet.inject();

var inserted = exports.inserted = {};

function values(cls, vars) {
  var hash = (0, _hash.hashArray)([cls].concat(vars));
  var varCls = 'vars-' + hash;
  if (inserted[hash]) {
    return varCls;
  }
  var src = vars.map(function (val, i) {
    return '--' + cls + '-' + i + ': ' + val;
  }).join('; ');
  sheet.insert('.' + varCls + ' {' + src + '}');
  inserted[hash] = true;

  return varCls;
}

function flush() {
  sheet.flush();
  exports.inserted = inserted = {};
  sheet.inject();
}

function css(classes, vars, content) {
  if (!Array.isArray(classes)) {
    classes = [classes];
  }

  var computedClassName = classes.map(function (cls) {
    return typeof cls === 'string' ? cls : objStyle(cls);
  }).join(' ').trim();

  if (content) {
    // inline mode
    var src = content.apply(undefined, vars); // returns an array
    var hash = (0, _hash.hashArray)(src);

    if (!inserted[hash]) {
      inserted[hash] = true;
      src.map(function (r) {
        return r.replace(new RegExp(classes[0], 'gm'), classes[0] + '-' + hash);
      }).forEach(function (r) {
        return sheet.insert(r);
      });
    }
    return classes[0] + '-' + hash + ' ' + computedClassName;
  }

  return computedClassName + (vars && vars.length > 0 ? ' ' + values(classes[0], vars) : '');
}

function injectGlobal(src) {
  var hash = (0, _hash.hashArray)(src);
  if (!inserted[hash]) {
    inserted[hash] = true;
    src.forEach(function (r) {
      return sheet.insert(r);
    });
  }
}

var fontFace = exports.fontFace = injectGlobal;

function keyframes(kfm, src) {
  var hash = (0, _hash.hashArray)(src);
  var animationName = kfm + '-' + hash;
  if (!inserted[hash]) {
    inserted[hash] = true;
    src.forEach(function (r) {
      sheet.insert('@keyframes ' + animationName + ' ' + r);
    });
  }
  return animationName;
}

function hydrate(ids) {
  ids.forEach(function (id) {
    return inserted[id] = true;
  });
}

// 🍩
// https://github.com/jxnblk/cxs/blob/master/src/monolithic/index.js
function objStyle(style) {
  var hash = (0, _hash.hashObject)(style);
  var className = 'css-' + hash;
  var selector = '.' + className;

  if (inserted[hash]) return className;

  var rules = deconstruct(selector, style);
  rules.forEach(function (rule) {
    return sheet.insert(rule);
  });

  inserted[hash] = true;

  return className;
}

function deconstruct(selector, styles, media) {
  var decs = [];
  var rules = [];

  var _loop = function _loop(key) {
    var value = styles[key];
    var type = typeof value === 'undefined' ? 'undefined' : _typeof(value);

    if (type === 'number' || type === 'string') {
      decs.push(createDec(key, value));
      return 'continue';
    } else if (Array.isArray(value)) {
      value.forEach(function (val) {
        decs.push(createDec(key, val));
      });
      return 'continue';
    } else if (/^:/.test(key)) {
      deconstruct(selector + key, value, media).forEach(function (r) {
        return rules.push(r);
      });
      return 'continue';
    } else if (/^@media/.test(key)) {
      deconstruct(selector, value, key).forEach(function (r) {
        return rules.push(r);
      });
      return 'continue';
    } else {
      deconstruct(selector + ' ' + key, value, media).forEach(function (r) {
        return rules.push(r);
      });
      return 'continue';
    }
  };

  for (var key in styles) {
    var _ret = _loop(key);

    if (_ret === 'continue') continue;
  }

  rules.unshift(createRule(selector, decs, media));

  return rules;
}

function createDec(key, value) {
  var prop = hyphenate(key);
  var val = addPx(key, value);
  return prop + ':' + val;
}

function createRule(selector, decs, media) {
  var rule = selector + '{' + decs.join(';') + '}';
  return media ? media + '{' + rule + '}' : rule;
}

function hyphenate(str) {
  return ('' + str).replace(/[A-Z]|^ms/g, '-$&').toLowerCase();
}

function addPx(prop, value) {
  if (typeof value !== 'number' || unitlessProps[prop]) return value;
  return value + 'px';
}

var unitlessProps = {
  animationIterationCount: 1,
  boxFlex: 1,
  boxFlexGroup: 1,
  boxOrdinalGroup: 1,
  columnCount: 1,
  flex: 1,
  flexGrow: 1,
  flexPositive: 1,
  flexShrink: 1,
  flexNegative: 1,
  flexOrder: 1,
  gridRow: 1,
  gridColumn: 1,
  fontWeight: 1,
  lineClamp: 1,
  lineHeight: 1,
  opacity: 1,
  order: 1,
  orphans: 1,
  tabSize: 1,
  widows: 1,
  zIndex: 1,
  zoom: 1,
  fillOpacity: 1,
  stopOpacity: 1,
  strokeDashoffset: 1,
  strokeOpacity: 1,
  strokeWidth: 1
};